@@ -1,4 +1,4 @@
-/*  Copyright (C) 2015 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
+/*  Copyright (C) 2015-2017 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -167,30 +167,41 @@ static int net_list(lua_State *L)
 	return 1;
 }
 
-/** Listen on interface address list. */
-static int net_listen_iface(lua_State *L, int port, int flags)
+/** Listen on an address list represented by the top of lua stack. */
+static int net_listen_addrs(lua_State *L, int port, int flags)
 {
-	/* Expand 'addr' key if exists */
-	lua_getfield(L, 1, "addr");
-	if (lua_isnil(L, -1)) {
+	/* Case: table with 'addr' field; only follow that field directly. */
+	lua_getfield(L, -1, "addr");
+	if (!lua_isnil(L, -1)) {
+		lua_replace(L, -2);
+	} else {
 		lua_pop(L, 1);
-		lua_pushvalue(L, 1);
 	}
 
-	/* Bind to address list */
-	struct engine *engine = engine_luaget(L);
-	size_t count = lua_rawlen(L, -1);
-	for (size_t i = 0; i < count; ++i) {
-		lua_rawgeti(L, -1, i + 1);
-		int ret = network_listen(&engine->net, lua_tostring(L, -1),
-		                         port, flags);
+	/* Case: string, representing a single address. */
+	const char *str = lua_tostring(L, -1);
+	if (str != NULL) {
+		struct engine *engine = engine_luaget(L);
+		int ret = network_listen(&engine->net, str, port, flags);
 		if (ret != 0) {
-			kr_log_info("[system] bind to '%s#%d' %s\n", lua_tostring(L, -1), port, kr_strerror(ret));
+			kr_log_info("[system] bind to '%s#%d' %s\n",
+					str, port, kr_strerror(ret));
 		}
-		lua_pop(L, 1);
+		return ret == 0;
 	}
 
-	lua_pushboolean(L, true);
+	/* Last case: table where all entries are added recursively. */
+	if (!lua_istable(L, -1)) {
+		format_error(L, "bad type for address");
+		lua_error(L);
+		return 0;
+	}
+	lua_pushnil(L);
+	while (lua_next(L, -2)) {
+		if (net_listen_addrs(L, port, flags) == 0)
+			return 0;
+		lua_pop(L, 1);
+	}
 	return 1;
 }
 
@@ -210,34 +221,28 @@ static int net_listen(lua_State *L)
 {
 	/* Check parameters */
 	int n = lua_gettop(L);
+	if (n < 1 || n > 3) {
+		format_error(L, "expected one to three arguments; usage:\n"
+			"net.listen(addressses, [port = 53, flags = {tls = (port == 853)}])\n");
+		lua_error(L);
+	}
+
 	int port = KR_DNS_PORT;
 	if (n > 1 && lua_isnumber(L, 2)) {
 		port = lua_tointeger(L, 2);
 	}
+
 	bool tls = (port == KR_DNS_TLS_PORT);
 	if (n > 2 && lua_istable(L, 3)) {
 		tls = table_get_flag(L, 3, "tls", tls);
 	}
-
-	/* Process interface or (address, port, flags) triple. */
 	int flags = tls ? (NET_TCP|NET_TLS) : (NET_TCP|NET_UDP);
-	if (lua_istable(L, 1)) {
-		return net_listen_iface(L, port, flags);
-	} else if (n < 1 || !lua_isstring(L, 1)) {
-		format_error(L, "expected 'listen(string addr, number port = 53[, bool tls = false])'");
-		lua_error(L);
-	}
-
-	/* Open resolution context cache */
-	struct engine *engine = engine_luaget(L);
-	int ret = network_listen(&engine->net, lua_tostring(L, 1), port, flags);
-	if (ret != 0) {
-		format_error(L, kr_strerror(ret));
-		lua_error(L);
-	}
-
-	lua_pushboolean(L, true);
-	return 1;
+	
+	/* Now focus on the first argument. */
+	lua_pop(L, n - 1);
+	int res = net_listen_addrs(L, port, flags);
+	lua_pushboolean(L, res);
+	return res;
 }
 
 /** Close endpoint. */
@@ -389,6 +394,33 @@ static int net_tls(lua_State *L)
 	return 1;
 }
 
+static int net_tls_padding(lua_State *L)
+{
+	struct engine *engine = engine_luaget(L);
+
+	/* Only return current padding. */
+	if (lua_gettop(L) == 0) {
+		if (engine->resolver.tls_padding == 0) {
+			return -1;
+		}
+		lua_pushinteger(L, engine->resolver.tls_padding);
+		return 1;
+	}
+
+	if ((lua_gettop(L) != 1) || !lua_isnumber(L, 1)) {
+		lua_pushstring(L, "net.tls_padding takes one numeric parameter: (\"padding\")");
+		lua_error(L);
+	}
+	int padding = lua_tointeger(L, 1);
+	if ((padding < 0) || (padding > MAX_TLS_PADDING)) {
+		lua_pushstring(L, "net.tls_padding parameter has to be a number between <0, 4096>");
+		lua_error(L);
+	}
+	engine->resolver.tls_padding = padding;
+	lua_pushboolean(L, true);
+	return 1;
+}
+
 int lib_net(lua_State *L)
 {
 	static const luaL_Reg lib[] = {
@@ -399,6 +431,7 @@ int lib_net(lua_State *L)
 		{ "bufsize",      net_bufsize },
 		{ "tcp_pipeline", net_pipeline },
 		{ "tls",          net_tls },
+		{ "tls_padding",  net_tls_padding },
 		{ NULL, NULL }
 	};
 	register_lib(L, "net", lib);
@@ -425,10 +458,11 @@ static int cache_count(lua_State *L)
 	struct engine *engine = engine_luaget(L);
 	const struct kr_cdb_api *api = engine->resolver.cache.api;
 
-	/* First key is a version counter, omit it. */
 	struct kr_cache *cache = &engine->resolver.cache;
-	if (kr_cache_is_open(cache)) {
-		lua_pushinteger(L, api->count(cache->db) - 1);
+	int count = api->count(cache->db);
+	if (kr_cache_is_open(cache) && count >= 0) {
+		/* First key is a version counter, omit it if nonempty. */
+		lua_pushinteger(L, count ? count - 1 : 0);
 		return 1;
 	}
 	return 0;
@@ -470,6 +504,53 @@ static const struct kr_cdb_api *cache_select(struct engine *engine, const char *
 	return NULL;
 }
 
+static int cache_max_ttl(lua_State *L)
+{
+	struct engine *engine = engine_luaget(L);
+	struct kr_cache *cache = &engine->resolver.cache;
+
+	int n = lua_gettop(L);
+	if (n > 0) {
+		if (!lua_isnumber(L, 1)) {
+			format_error(L, "expected 'max_ttl(number ttl)'");
+			lua_error(L);
+		}
+		uint32_t min = cache->ttl_min;
+		int64_t ttl = lua_tonumber(L, 1);
+		if (ttl < 0 || ttl <= min || ttl > UINT32_MAX) {
+			format_error(L, "max_ttl must be larger than minimum TTL, and in range <1, UINT32_MAX>'");
+			lua_error(L);
+		}
+		cache->ttl_max = ttl;
+	}
+	lua_pushinteger(L, cache->ttl_max);
+	return 1;
+}
+
+
+static int cache_min_ttl(lua_State *L)
+{
+	struct engine *engine = engine_luaget(L);
+	struct kr_cache *cache = &engine->resolver.cache;
+
+	int n = lua_gettop(L);
+	if (n > 0) {
+		if (!lua_isnumber(L, 1)) {
+			format_error(L, "expected 'min_ttl(number ttl)'");
+			lua_error(L);
+		}
+		uint32_t max = cache->ttl_max;
+		int64_t ttl = lua_tonumber(L, 1);
+		if (ttl < 0 || ttl >= max || ttl > UINT32_MAX) {
+			format_error(L, "min_ttl must be smaller than maximum TTL, and in range <0, UINT32_MAX>'");
+			lua_error(L);
+		}
+		cache->ttl_min = ttl;
+	}
+	lua_pushinteger(L, cache->ttl_min);
+	return 1;
+}
+
 /** Open cache */
 static int cache_open(lua_State *L)
 {
@@ -653,12 +734,9 @@ static int cache_clear(lua_State *L)
 	}
 
 	/* Clear reputation tables */
-	lru_deinit(engine->resolver.cache_rtt);
-	lru_deinit(engine->resolver.cache_rep);
-	lru_init(engine->resolver.cache_rtt, LRU_RTT_SIZE);
-	lru_init(engine->resolver.cache_rep, LRU_REP_SIZE);
-	lru_deinit(engine->resolver.cache_cookie);
-	lru_init(engine->resolver.cache_cookie, LRU_COOKIES_SIZE);
+	lru_reset(engine->resolver.cache_rtt);
+	lru_reset(engine->resolver.cache_rep);
+	lru_reset(engine->resolver.cache_cookie);
 	lua_pushboolean(L, true);
 	return 1;
 }
@@ -745,6 +823,8 @@ int lib_cache(lua_State *L)
 		{ "prune",  cache_prune },
 		{ "clear",  cache_clear },
 		{ "get",    cache_get },
+		{ "max_ttl", cache_max_ttl },
+		{ "min_ttl", cache_min_ttl },
 		{ NULL, NULL }
 	};
 
