@@ -232,6 +232,15 @@ static int l_option(lua_State *L)
 	return 1;
 }
 
+/** @internal for l_trustanchor: */
+static void ta_add(zs_scanner_t *zs)
+{
+	map_t *ta = zs->process.data;
+	if (!ta)
+		return;
+	if (kr_ta_add(ta, zs->r_owner, zs->r_type, zs->r_ttl, zs->r_data, zs->r_data_length))
+		zs->process.data = NULL; /* error signalling */
+}
 /** Enable/disable trust anchor. */
 static int l_trustanchor(lua_State *L)
 {
@@ -260,13 +269,11 @@ static int l_trustanchor(lua_State *L)
 		lua_pushstring(L, "not enough memory");
 		lua_error(L);
 	}
-	int ok = zs_set_input_string(zs, anchor, strlen(anchor)) == 0 &&
-	         zs_parse_all(zs) == 0;
-	/* Add it to TA set and cleanup */
-	if (ok) {
-		ok = kr_ta_add(&engine->resolver.trust_anchors,
-		               zs->r_owner, zs->r_type, zs->r_ttl, zs->r_data, zs->r_data_length) == 0;
-	}
+	zs_set_processing(zs, ta_add, NULL, &engine->resolver.trust_anchors);
+	bool ok = zs_set_input_string(zs, anchor, strlen(anchor)) == 0
+		&& zs_parse_all(zs) == 0;
+	ok = ok && zs->process.data; /* reset to NULL on error in ta_add */
+
 	zs_deinit(zs);
 	free(zs);
 	/* Report errors */
@@ -376,6 +383,11 @@ static int l_tojson(lua_State *L)
 
 static int l_map(lua_State *L)
 {
+	if (lua_gettop(L) != 1 || !lua_isstring(L, 1)) {
+		lua_pushliteral(L, "map('string with a lua expression')");
+		lua_error(L);
+	}
+
 	struct engine *engine = engine_luaget(L);
 	const char *cmd = lua_tostring(L, 1);
 	uint32_t len = strlen(cmd);
@@ -538,8 +550,6 @@ static int init_state(struct engine *engine)
 	lua_setglobal(engine->L, "map");
 	lua_pushliteral(engine->L, MODULEDIR);
 	lua_setglobal(engine->L, "moduledir");
-	lua_pushliteral(engine->L, ETCDIR);
-	lua_setglobal(engine->L, "etcdir");
 	lua_pushlightuserdata(engine->L, engine);
 	lua_setglobal(engine->L, "__engine");
 	return kr_ok();
