@@ -19,6 +19,9 @@
 #include <uv.h>
 #include <gnutls/gnutls.h>
 #include <libknot/packet/pkt.h>
+#include "lib/defines.h"
+
+#define MAX_TLS_PADDING KR_EDNS_PAYLOAD
 
 struct tls_ctx_t;
 struct tls_credentials;
@@ -27,6 +30,8 @@ struct tls_credentials {
 	char *tls_cert;
 	char *tls_key;
 	gnutls_certificate_credentials_t credentials;
+	time_t valid_until;
+	char *ephemeral_servicename;
 };
 
 /*! Toggle verbose logging from TLS context. */
@@ -41,7 +46,9 @@ void tls_free(struct tls_ctx_t* tls);
 /*! Push new data to TLS context for sending */
 int tls_push(struct qr_task *task, uv_handle_t* handle, knot_pkt_t * pkt);
 
-/*! Unwrap incoming data from a TLS stream and pass them to TCP session. */
+/*! Unwrap incoming data from a TLS stream and pass them to TCP session.
+ * @return the number of newly-completed requests (>=0) or an error code
+ */
 int tls_process(struct worker_ctx *worker, uv_stream_t *handle, const uint8_t *buf, ssize_t nread);
 
 /*! Set TLS certificate and key from files. */
@@ -59,3 +66,6 @@ void tls_credentials_free(struct tls_credentials *tls_credentials);
 /*! Log DNS-over-TLS OOB key-pin form of current credentials:
  * https://tools.ietf.org/html/rfc7858#appendix-A */
 void tls_credentials_log_pins(struct tls_credentials *tls_credentials);
+
+/*! Generate new ephemeral TLS credentials. */
+struct tls_credentials * tls_get_ephemeral_credentials(struct engine *engine);
