@@ -1,4 +1,4 @@
-/*  Copyright (C) 2014 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
+/*  Copyright (C) 2014-2017 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -25,13 +25,18 @@
 #include <malloc.h>
 #endif
 #include <assert.h>
+#include <sys/types.h>
+#include <unistd.h>
 #include "lib/utils.h"
 #include "lib/layer.h"
 #include "daemon/worker.h"
+#include "daemon/bindings.h"
 #include "daemon/engine.h"
 #include "daemon/io.h"
 #include "daemon/tls.h"
 
+#define VERBOSE_MSG(qry, fmt...) QRVERBOSE(qry, "wrkr", fmt)
+
 /* @internal Union of various libuv objects for freelist. */
 struct req
 {
@@ -48,13 +53,6 @@ struct req
 	} as;
 };
 
-/** @internal Debugging facility. */
-#ifdef DEBUG
-#define DEBUG_MSG(fmt...) printf("[daem] " fmt)
-#else
-#define DEBUG_MSG(fmt...)
-#endif
-
 /* Convenience macros */
 #define qr_task_ref(task) \
 	do { ++(task)->refs; } while(0)
@@ -256,6 +254,8 @@ static struct qr_task *qr_task_create(struct worker_ctx *worker, uv_handle_t *ha
 	task->req.qsource.key = NULL;
 	task->req.qsource.addr = NULL;
 	task->req.qsource.dst_addr = NULL;
+	task->req.qsource.packet = NULL;
+	task->req.qsource.opt = NULL;
 	/* Remember query source addr */
 	if (addr) {
 		size_t addr_len = sizeof(struct sockaddr_in);
@@ -483,7 +483,6 @@ static int qr_task_send(struct qr_task *task, uv_handle_t *handle, struct sockad
 		}
 	}
 	/* Send using given protocol */
-	ret = 0;
 	if (handle->type == UV_UDP) {
 		uv_buf_t buf = { (char *)pkt->wire, pkt->size };
 		send_req->as.send.data = task;
@@ -524,7 +523,10 @@ static void on_connect(uv_connect_t *req, int status)
 	uv_stream_t *handle = req->handle;
 	if (qr_valid_handle(task, (uv_handle_t *)req->handle)) {
 		if (status == 0) {
-			qr_task_send(task, (uv_handle_t *)handle, task->addrlist, task->pktbuf);
+			struct sockaddr_storage addr;
+			int addr_len = sizeof(addr);
+			uv_tcp_getpeername((uv_tcp_t *)handle, (struct sockaddr *)&addr, &addr_len);
+			qr_task_send(task, (uv_handle_t *)handle, (struct sockaddr *)&addr, task->pktbuf);
 		} else {
 			qr_task_step(task, task->addrlist, NULL);
 		}
@@ -552,10 +554,10 @@ static void on_timeout(uv_timer_t *req)
 		struct sockaddr_in6 *addrlist = (struct sockaddr_in6 *)task->addrlist;
 		for (uint16_t i = 0; i < MIN(task->pending_count, task->addrlist_count); ++i) {
 			struct sockaddr *choice = (struct sockaddr *)(&addrlist[i]);
-			WITH_DEBUG {
+			WITH_VERBOSE {
 				char addr_str[INET6_ADDRSTRLEN];
 				inet_ntop(choice->sa_family, kr_inaddr(choice), addr_str, sizeof(addr_str));
-				QRDEBUG(qry, "wrkr", "=> server: '%s' flagged as 'bad'\n", addr_str);
+				VERBOSE_MSG(qry, "=> server: '%s' flagged as 'bad'\n", addr_str);
 			}
 			kr_nsrep_update_rtt(&qry->ns, choice, KR_NS_TIMEOUT,
 					    worker->engine->resolver.cache_rtt, KR_NS_UPDATE);
@@ -697,7 +699,7 @@ static int qr_task_finalize(struct qr_task *task, int state)
 	task->finished = true;
 	/* Send back answer */
 	(void) qr_task_send(task, task->source.handle, (struct sockaddr *)&task->source.addr, task->req.answer);
-	return state == KNOT_STATE_DONE ? 0 : kr_error(EIO);
+	return state == KR_STATE_DONE ? 0 : kr_error(EIO);
 }
 
 static int qr_task_step(struct qr_task *task, const struct sockaddr *packet_source, knot_pkt_t *packet)
@@ -713,16 +715,17 @@ static int qr_task_step(struct qr_task *task, const struct sockaddr *packet_sour
 	task->addrlist = NULL;
 	task->addrlist_count = 0;
 	task->addrlist_turn = 0;
+	task->req.has_tls = (task->session && task->session->has_tls);
 	int state = kr_resolve_consume(&task->req, packet_source, packet);
-	while (state == KNOT_STATE_PRODUCE) {
+	while (state == KR_STATE_PRODUCE) {
 		state = kr_resolve_produce(&task->req, &task->addrlist, &sock_type, task->pktbuf);
 		if (unlikely(++task->iter_count > KR_ITER_LIMIT || task->timeouts >= KR_TIMEOUT_LIMIT)) {
-			return qr_task_finalize(task, KNOT_STATE_FAIL);
+			return qr_task_finalize(task, KR_STATE_FAIL);
 		}
 	}
 
 	/* We're done, no more iterations needed */
-	if (state & (KNOT_STATE_DONE|KNOT_STATE_FAIL)) {
+	if (state & (KR_STATE_DONE|KR_STATE_FAIL)) {
 		return qr_task_finalize(task, state);
 	} else if (!task->addrlist || sock_type < 0) {
 		return qr_task_step(task, NULL, NULL);
@@ -742,9 +745,21 @@ static int qr_task_step(struct qr_task *task, const struct sockaddr *packet_sour
 		if (subreq_enqueue(task)) {
 			return kr_ok(); /* Will be notified when outgoing query finishes. */
 		}
+		/* Check current query NSLIST */
+		struct kr_query *qry = array_tail(task->req.rplan.pending);
 		/* Start transmitting */
 		if (retransmit(task)) {
-			ret = timer_start(task, on_retransmit, KR_CONN_RETRY, 0);
+			assert(qry != NULL);
+			/* Retransmit at default interval, or more frequently if the mean
+			 * RTT of the server is better. If the server is glued, use default rate. */
+			size_t timeout = qry->ns.score;
+			if (timeout > KR_NS_GLUED) {
+				/* We don't have information about variance in RTT, expect +10ms */
+				timeout = MIN(qry->ns.score + 10, KR_CONN_RETRY);
+			} else {
+				timeout = KR_CONN_RETRY;
+			}
+			ret = timer_start(task, on_retransmit, timeout, 0);
 		} else {
 			return qr_task_step(task, NULL, NULL);
 		}
@@ -763,7 +778,7 @@ static int qr_task_step(struct qr_task *task, const struct sockaddr *packet_sour
 			return qr_task_step(task, NULL, NULL);
 		}
 		conn->data = task;
-		if (uv_tcp_connect(conn, (uv_tcp_t *)client, task->addrlist, on_connect) != 0) {
+		if (uv_tcp_connect(conn, (uv_tcp_t *)client, packet_source?packet_source:task->addrlist, on_connect) != 0) {
 			req_release(task->worker, (struct req *)conn);
 			return qr_task_step(task, NULL, NULL);
 		}
@@ -774,7 +789,7 @@ static int qr_task_step(struct qr_task *task, const struct sockaddr *packet_sour
 	/* Start next step with timeout, fatal if can't start a timer. */
 	if (ret != 0) {
 		subreq_finalize(task, packet_source, packet);
-		return qr_task_finalize(task, KNOT_STATE_FAIL);
+		return qr_task_finalize(task, KR_STATE_FAIL);
 	}
 	return 0;
 }
@@ -861,7 +876,7 @@ int worker_end_tcp(struct worker_ctx *worker, uv_handle_t *handle)
 	 * borrowed the task from parent session. */
 	struct session *session = handle->data;
 	if (session->outgoing) {
-		worker_submit(worker, (uv_handle_t *)handle, NULL, NULL);	
+		worker_submit(worker, (uv_handle_t *)handle, NULL, NULL);
 	} else {
 		discard_buffered(session);
 	}
@@ -899,7 +914,12 @@ int worker_process_tcp(struct worker_ctx *worker, uv_stream_t *handle, const uin
 	 * to buffer incoming message until it's complete. */
 	if (!session->outgoing) {
 		if (!task) {
-			task = qr_task_create(worker, (uv_handle_t *)handle, NULL);
+			/* Get TCP peer name, keep zeroed address if it fails. */
+			struct sockaddr_storage addr;
+			memset(&addr, 0, sizeof(addr));
+			int addr_len = sizeof(addr);
+			uv_tcp_getpeername((uv_tcp_t *)handle, (struct sockaddr *)&addr, &addr_len);
+			task = qr_task_create(worker, (uv_handle_t *)handle, (struct sockaddr *)&addr);
 			if (!task) {
 				return kr_error(ENOMEM);
 			}
@@ -1007,7 +1027,8 @@ int worker_resolve(struct worker_ctx *worker, knot_pkt_t *query, unsigned option
 	return qr_task_step(task, NULL, query);
 }
 
-int worker_reserve(struct worker_ctx *worker, size_t ring_maxlen)
+/** Reserve worker buffers */
+static int worker_reserve(struct worker_ctx *worker, size_t ring_maxlen)
 {
 	array_init(worker->pool_mp);
 	array_init(worker->pool_ioreq);
@@ -1042,4 +1063,38 @@ void worker_reclaim(struct worker_ctx *worker)
 	map_clear(&worker->outgoing);
 }
 
-#undef DEBUG_MSG
+struct worker_ctx *worker_create(struct engine *engine, knot_mm_t *pool,
+		int worker_id, int worker_count)
+{
+	/* Load bindings */
+	engine_lualib(engine, "modules", lib_modules);
+	engine_lualib(engine, "net",     lib_net);
+	engine_lualib(engine, "cache",   lib_cache);
+	engine_lualib(engine, "event",   lib_event);
+	engine_lualib(engine, "worker",  lib_worker);
+
+	/* Create main worker. */
+	struct worker_ctx *worker = mm_alloc(pool, sizeof(*worker));
+	if (!worker) {
+		return NULL;
+	}
+	memset(worker, 0, sizeof(*worker));
+	worker->id = worker_id;
+	worker->count = worker_count;
+	worker->engine = engine;
+	worker_reserve(worker, MP_FREELIST_SIZE);
+	/* Register worker in Lua thread */
+	lua_pushlightuserdata(engine->L, worker);
+	lua_setglobal(engine->L, "__worker");
+	lua_getglobal(engine->L, "worker");
+	lua_pushnumber(engine->L, worker_id);
+	lua_setfield(engine->L, -2, "id");
+	lua_pushnumber(engine->L, getpid());
+	lua_setfield(engine->L, -2, "pid");
+	lua_pushnumber(engine->L, worker_count);
+	lua_setfield(engine->L, -2, "count");
+	lua_pop(engine->L, 1);
+	return worker;
+}
+
+#undef VERBOSE_MSG
