@@ -1,4 +1,4 @@
-/*  Copyright (C) 2015 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
+/*  Copyright (C) 2015-2017 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -17,6 +17,7 @@
 #pragma once
 
 #include "lib/defines.h"
+#include "lib/utils.h"
 #include <libknot/packet/pkt.h>
 
 /**
@@ -46,12 +47,15 @@ struct dseckey;
 /** DNSSEC validation context. */
 struct kr_rrset_validation_ctx {
 	const knot_pkt_t *pkt;		/*!< Packet to be validated. */
+	ranked_rr_array_t *rrs;		/*!< List of preselected RRs to be validated. */
 	knot_section_t section_id;	/*!< Section to work with. */
 	const knot_rrset_t *keys;	/*!< DNSKEY RRSet. */
         const knot_dname_t *zone_name;	/*!< Name of the zone containing the RRSIG RRSet. */
 	uint32_t timestamp;		/*!< Validation time. */
         bool has_nsec3;			/*!< Whether to use NSEC3 validation. */
+	uint32_t qry_uid;		/*!< Current query uid. */
 	uint32_t flags;			/*!< Output - Flags. */
+	uint32_t err_cnt;		/*!< Output - Number of validation failures. */
 	int result;			/*!< Output - 0 or error code. */
 };
 
