@@ -1,4 +1,4 @@
-/*  Copyright (C) 2015 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
+/*  Copyright (C) 2015-2017 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -24,10 +24,10 @@
 #include "lib/module.h"
 
 /* List of embedded modules */
-const knot_layer_api_t *iterate_layer(struct kr_module *module);
-const knot_layer_api_t *validate_layer(struct kr_module *module);
-const knot_layer_api_t *rrcache_layer(struct kr_module *module);
-const knot_layer_api_t *pktcache_layer(struct kr_module *module);
+const kr_layer_api_t *iterate_layer(struct kr_module *module);
+const kr_layer_api_t *validate_layer(struct kr_module *module);
+const kr_layer_api_t *rrcache_layer(struct kr_module *module);
+const kr_layer_api_t *pktcache_layer(struct kr_module *module);
 static const struct kr_module embedded_modules[] = {
 	{ "iterate",  NULL, NULL, NULL, iterate_layer, NULL, NULL, NULL },
 	{ "validate", NULL, NULL, NULL, validate_layer, NULL, NULL, NULL },
