@@ -1,4 +1,4 @@
-/*  Copyright (C) 2014 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
+/*  Copyright (C) 2014-2017 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -17,6 +17,7 @@
 #pragma once
 
 #include <netinet/in.h>
+#include <sys/socket.h>
 #include <libknot/packet/pkt.h>
 
 #include "lib/cookies/control.h"
@@ -51,11 +52,11 @@
  * state = kr_resolve_consume(&req, query);
  *
  * // Generate answer
- * while (state == KNOT_STATE_PRODUCE) {
+ * while (state == KR_STATE_PRODUCE) {
  *
  *     // Additional query generate, do the I/O and pass back answer
  *     state = kr_resolve_produce(&req, &addr, &type, query);
- *     while (state == KNOT_STATE_CONSUME) {
+ *     while (state == KR_STATE_CONSUME) {
  *         int ret = sendrecv(addr, proto, query, resp);
  *
  *         // If I/O fails, make "resp" empty
@@ -71,6 +72,16 @@
  * @endcode
  */
 
+/** Validation rank */
+typedef enum kr_validation_rank {
+	KR_VLDRANK_INITIAL   = 0,   /* Entry was just added; not validated yet. */
+	KR_VLDRANK_INSECURE  = 1,   /* Entry is DNSSEC insecure (e.g. RRSIG not exists). */
+	KR_VLDRANK_BAD	     = 2,   /* Matching RRSIG found, but validation fails. */
+	KR_VLDRANK_MISMATCH  = 3,   /* RRSIG signer name is */
+	KR_VLDRANK_UNKNOWN   = 4,   /* Unknown */
+	KR_VLDRANK_SECURE    = 5    /* Entry is DNSSEC valid (e.g. RRSIG exists). */
+} kr_validation_rank_t;
+
 /** @cond internal Array of modules. */
 typedef array_t(struct kr_module *) module_array_t;
 /* @endcond */
@@ -98,6 +109,7 @@ struct kr_context
 	 * module because of better access. */
 	struct kr_cookie_ctx cookie_ctx;
 	kr_cookie_lru_t *cache_cookie;
+	uint32_t tls_padding; /**< See net.tls_padding in ../daemon/README.rst */
 	knot_mm_t *pool;
 };
 
@@ -112,26 +124,30 @@ struct kr_context
  * @note All data for this request must be allocated from the given pool.
  */
 struct kr_request {
-    struct kr_context *ctx;
-    knot_pkt_t *answer;
-    struct kr_query *current_query;    /**< Current evaluated query. */
-    struct {
-        const knot_rrset_t *key;
-        const struct sockaddr *addr;
-        const struct sockaddr *dst_addr;
-        const knot_pkt_t *packet;
-        const knot_rrset_t *opt;
-    } qsource;
-    struct {
-        unsigned rtt;                  /**< Current upstream RTT */
-        const struct sockaddr *addr;   /**< Current upstream address */
-    } upstream;                        /**< Upstream information, valid only in consume() phase */
-    uint32_t options;
-    int state;
-    rr_array_t authority;
-    rr_array_t additional;
-    struct kr_rplan rplan;
-    knot_mm_t pool;
+	struct kr_context *ctx;
+	knot_pkt_t *answer;
+	struct kr_query *current_query;    /**< Current evaluated query. */
+	struct {
+		const knot_rrset_t *key;
+		const struct sockaddr *addr;
+		const struct sockaddr *dst_addr;
+		const knot_pkt_t *packet;
+		const knot_rrset_t *opt;
+	} qsource;
+	struct {
+		unsigned rtt;                  /**< Current upstream RTT */
+		const struct sockaddr *addr;   /**< Current upstream address */
+	} upstream;                        /**< Upstream information, valid only in consume() phase */
+	uint32_t options;
+	int state;
+	ranked_rr_array_t answ_selected;
+	ranked_rr_array_t auth_selected;
+	rr_array_t additional;
+	bool answ_validated;
+	bool auth_validated;
+	struct kr_rplan rplan;
+	int has_tls;
+	knot_mm_t pool;
 };
 
 /**
