@@ -1,4 +1,6 @@
 local kres = require('kres')
+local bit = require('bit')
+local ffi = require('ffi')
 
 -- Counter of unique rules
 local nextid = 0
@@ -29,15 +31,20 @@ if has_socket then
 		return s
 	end
 end
-local has_ffi, ffi = pcall(require, 'ffi')
-if not has_ffi then
-	socket_client = function () return error("missing ffi library, required for this policy") end
+
+local function parse_target(target)
+	local addr, port = target:match '([^@]*)@?(.*)'
+	port = port and tonumber(port) or 53
+	addr = kres.str2ip(addr)
+	if addr == nil then
+		error("target '"..target..'" is not a valid IP address')
+	end
+	return addr, port
 end
 
 -- Mirror request elsewhere, and continue solving
 local function mirror(target)
-	local addr, port = target:match '([^@]*)@?(.*)'
-	if not port or #port == 0 then port = 53 end
+	local addr, port = parse_target(target)
 	local sink, err = socket_client(addr, port)
 	if not sink then panic('MIRROR target %s is not a valid: %s', target, err) end
 	return function(state, req)
@@ -53,13 +60,21 @@ end
 
 -- Forward request, and solve as stub query
 local function forward(target)
-	local dst_ip = kres.str2ip(target)
-	if dst_ip == nil then error("FORWARD target '"..target..'" is not a valid IP address') end
+	local list = {}
+	if type(target) == 'table' then
+		for _, v in pairs(target) do
+			table.insert(list, {parse_target(v)})
+			assert(#list <= 4, 'at most 4 FORWARD targets are supported')
+		end
+	else
+		table.insert(list, {parse_target(target)})
+	end
 	return function(state, req)
 		req = kres.request_t(req)
 		local qry = req:current()
-		qry.flags = qry.flags + kres.query.STUB
-		qry:nslist(dst_ip)
+		-- Switch mode to stub resolver, do not track origin zone cut since it's not real authority NS
+		qry.flags = bit.band(bit.bor(qry.flags, kres.query.STUB), bit.bnot(kres.query.ALWAYS_CUT))
+		qry:nslist(list)
 		return state
 	end
 end
@@ -73,13 +88,12 @@ local function reroute(tbl, names)
 		table.insert(prefixes, names and ren.name(from, to) or ren.prefix(from, to))
 	end
 	-- Return rule closure
-	tbl = nil
 	return ren.rule(prefixes)
 end
 
 local policy = {
 	-- Policies
-	PASS = 1, DENY = 2, DROP = 3, TC = 4, FORWARD = forward, REROUTE = reroute, MIRROR = mirror,
+	PASS = 1, DENY = 2, DROP = 3, TC = 4, QTRACE = 5, FORWARD = forward, REROUTE = reroute, MIRROR = mirror,
 	-- Special values
 	ANY = 0,
 }
@@ -135,7 +149,6 @@ end
 
 local function rpz_parse(action, path)
 	local rules = {}
-	local ffi = require('ffi')
 	local action_map = {
 		-- RPZ Policy Actions
 		['\0'] = action,
@@ -215,6 +228,10 @@ function policy.enforce(state, req, action)
 			answer:tc(1) -- ^ Only UDP queries
 			return kres.DONE
 		end
+	elseif action == policy.QTRACE then
+		local qry = req:current()
+		req.options = bit.bor(req.options, kres.query.TRACE)
+		qry.flags = bit.bor(qry.flags, kres.query.TRACE)
 	elseif type(action) == 'function' then
 		return action(state, req)
 	end
